SystemOrganization addCategory: #Breakout!

PasteUpMorph subclass: #BOPlayfield
	instanceVariableNames: 'stones ball bar keyActions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Breakout'!

!BOPlayfield commentStamp: 'tfel 11/22/2009 16:22' prior: 0!
I am the playfield responsible for adding my submorphs (a ball, a bar and a few stones). I pretty much let them do whatever they want, delegating keys to the bar as I think plausible and telling the ball whether or not he hits something (like my wall or one of my submorphs) at a specific point.
I also draw a game over screen on myself, if the game is won or lost.

Instance Variables
	ball:		hopefully a BoBall
	bar:		should be some bar that answers (at the very least) to moveLeft and moveRight
	keyActions:		a dictionary with keys I react on - configuration on the class side, please
	stones:	a collection of stones

ball
	- xxxxx

bar
	- xxxxx

keyActions
	- xxxxx

stones
	- xxxxx
!

!BOPlayfield class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:35'!
backgroundColor

	^ Color black! !

!BOPlayfield class methodsFor: 'parts bin' stamp: 'tfel 11/22/2009 16:10'!
descriptionForPartsBin
	^ self partName:	'SWA2009 Tutorial Breakout'
		categories:		#('Games')
		documentation:	'A simple Breakout-ish game from the SWA tutorial 2009'! !

!BOPlayfield class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:33'!
extent

	^ 640@480! !

!BOPlayfield class methodsFor: 'configuration' stamp: 'tfel 11/19/2009 09:19'!
keys

	^ Dictionary new
		add: #quit -> $q;
		add: #left -> Character arrowLeft;
		add: #right -> Character arrowRight;
		add: #start -> $p;
		add: #reset -> $r;
		yourself! !

!BOPlayfield class methodsFor: 'levels' stamp: 'tfel 11/19/2009 09:18'!
level

	^ Dictionary
		with: #positions -> {1@1 . 2@1 . 3@1 . 4@1 . 5@1 .
									1@2 .       3@2 .
							      2@3 .       4@3}
		with: #stoneExtent -> (self extent / (7@20))! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/6/2009 17:40'!
ball
	^ ball! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/19/2009 09:30'!
ball: aBall

	ball := aBall.! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/6/2009 17:40'!
bar
	^ bar! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/19/2009 09:30'!
bar: aBar

	bar := aBar.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/19/2009 09:24'!
createStonesWithFraction: aPoint 

	^ ((self class level at: #positions) collect: [:position | 
		BOStone 
			newAt: position
			withExtent: aPoint ]) asOrderedCollection! !

!BOPlayfield methodsFor: 'end-game' stamp: 'tfel 11/18/2009 09:21'!
gameOverScreenWith: aString
	
	^ RectangleMorph new
		color: (Color lightGray alpha: 0.5);
		bounds: self bounds;
		addMorph: ((StringMorph contents: aString)
							center: self center;
							color: Color red;
							yourself);
		yourself.! !

!BOPlayfield methodsFor: 'events-processing' stamp: 'tfel 11/18/2009 15:11'!
handleHitStone: aStone

	aStone ifNil: [^ false].
	aStone takeHit.
	self stones 
			remove: aStone;
			ifEmpty: [self win].
	^ true! !

!BOPlayfield methodsFor: 'events-processing' stamp: 'tfel 11/18/2009 09:30'!
handleKeyDown: anEvent

	(self keyActions 
		at: anEvent keyCharacter
		ifAbsent: []) value.! !

!BOPlayfield methodsFor: 'geometry testing' stamp: 'tfel 11/22/2009 16:07'!
hitsStoneAt: aPoint
	
	^ self handleHitStone: (self stones
										detect: [:stone | stone containsPoint: aPoint]
										ifNone: [nil])! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/22/2009 16:12'!
initialize

	super initialize.
	self 
		openInWorld;
		initializeKeys;
		initializeView;
		initializeSubmorphs.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/19/2009 09:33'!
initializeBall

	self ball ifNotNil: [self removeMorph: self ball].
	self 
		ball: (BOBall new playfield: self);
		addMorph: self ball.
	! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/19/2009 09:23'!
initializeBar

	self ball ifNotNil: [self removeMorph: self ball].
	self 
		bar: BOBar new;
		addMorph: self bar.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/18/2009 09:30'!
initializeKeys

	keyActions := Dictionary new
		at: (self class keys at: #quit) put: [self abandon];
		at: (self class keys at: #left) put: [self bar moveLeft];
		at: (self class keys at: #right) put: [self bar moveRight];
		at: (self class keys at: #start) put: [self ball startStepping];
		at: (self class keys at: #reset) put: [self reset];
		yourself.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/19/2009 09:27'!
initializeStones

	self stones ifNotNil: [self removeAllMorphsIn: self stones].
	self 
		stones: (self createStonesWithFraction: (self class level at: #stoneExtent));
		addAllMorphs: self stones.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/19/2009 09:33'!
initializeSubmorphs

	self
		initializeBar;
		initializeStones;
		initializeBall;
		stopSteppingSelfAndSubmorphs.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/22/2009 16:12'!
initializeToStandAlone

	self initialize.! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/18/2009 09:24'!
initializeView

	self
		extent: self class extent;
		color: self class backgroundColor.! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/18/2009 09:11'!
keyActions

	^ keyActions! !

!BOPlayfield methodsFor: 'end-game' stamp: 'tfel 11/18/2009 09:24'!
loose

	self
		addMorph: (self gameOverScreenWith: 'YOU LOOSE!!');
		stopSteppingSelfAndSubmorphs.! !

!BOPlayfield methodsFor: 'geometry testing' stamp: 'tfel 11/18/2009 09:25'!
reflectsBallAt: aPoint 

	self bottom < aPoint y ifTrue: [self loose].
	^ (self containsPoint: aPoint) not 
		or: [ self bar containsPoint: aPoint ]
		or: [ self hitsStoneAt: aPoint].! !

!BOPlayfield methodsFor: 'initialization' stamp: 'tfel 11/18/2009 09:25'!
reset
	"Reset all morphs: faster than making a new playfield"

	self
		removeAllMorphs;
		initializeSubmorphs.! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/6/2009 17:40'!
stones
	^ stones! !

!BOPlayfield methodsFor: 'accessing' stamp: 'tfel 11/19/2009 09:31'!
stones: aCollection

	stones := aCollection.! !

!BOPlayfield methodsFor: 'end-game' stamp: 'tfel 11/18/2009 09:25'!
win

	self
		addMorph: (self gameOverScreenWith: 'YOU ... win');
		stopSteppingSelfAndSubmorphs.! !

CircleMorph subclass: #BOBall
	instanceVariableNames: 'direction speed playfield'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Breakout'!

!BOBall commentStamp: 'tfel 11/22/2009 16:24' prior: 0!
I can be added to a playfield and I will try to fly around in it, using the playfield's 'reflectsBallAt:' method to tell me whether I have to turn around.

Instance Variables
	direction:		The direction I'm flying in. Just a point.
	playfield:		My actual playfield, that should respond to 'reflectsBallAt:'
	speed:			The distance I travel with each step.
!

!BOBall class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:17'!
color

	^ Color white duller! !

!BOBall class methodsFor: 'accessing' stamp: 'tfel 11/18/2009 15:28'!
directions
	^ { 0@1. 1@0. 0@-1. -1@0 }! !

!BOBall class methodsFor: 'configuration' stamp: 'tfel 11/19/2009 08:45'!
extentFraction
	
	^ 20! !

!BOBall class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:58'!
initialDirection

	^ {1@-1 . -1@-1} atRandom! !

!BOBall class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:12'!
speed

	^ 10 atRandom + 5! !

!BOBall methodsFor: 'stepping and presenter' stamp: 'tfel 11/22/2009 16:06'!
bounceOff

	self class directions do: [ :aDirection | 
		(self playfield 
			reflectsBallAt: (self center translateBy: self radius * aDirection)) 
		ifTrue: [self reflectIn: aDirection]].! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/6/2009 18:22'!
direction
	^ direction! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/18/2009 09:20'!
direction: aPoint
	direction := aPoint.! !

!BOBall methodsFor: 'initialization' stamp: 'tfel 11/19/2009 08:57'!
initialize
	
	super initialize.
	self
		direction: self class initialDirection;
		speed: self class speed;
		color: self class color! !

!BOBall methodsFor: 'submorphs-accessing' stamp: 'tfel 11/22/2009 16:04'!
noteNewOwner: aMorph

	aMorph isPlayfieldLike ifFalse: [^ self].
	self
		extent: aMorph extent / self class extentFraction;
		center: aMorph center.! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/18/2009 08:14'!
playfield
	^ playfield! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/19/2009 08:46'!
playfield: aPlayfield
	playfield := aPlayfield.! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/19/2009 08:46'!
radius

	^ self width / 2.! !

!BOBall methodsFor: 'transforming' stamp: 'tfel 11/19/2009 09:15'!
reflectIn: aDirection

	self direction: (self direction reflectedAbout: aDirection) negated! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/6/2009 18:22'!
speed
	^ speed! !

!BOBall methodsFor: 'accessing' stamp: 'tfel 11/9/2009 16:29'!
speed: aNumber
	speed := aNumber.! !

!BOBall methodsFor: 'stepping and presenter' stamp: 'tfel 11/18/2009 09:20'!
step
	"Let the ball bounce around in the world"
	self center: self center + (self direction * self speed).
	self bounceOff.! !

!BOBall methodsFor: 'testing' stamp: 'tfel 11/6/2009 18:50'!
stepTime

	^ 40! !

RectangleMorph subclass: #BOBar
	instanceVariableNames: 'speed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Breakout'!

!BOBar commentStamp: 'tfel 11/22/2009 16:22' prior: 0!
I can be moved using moveLeft and moveRight.

Instance Variables
	speed:	The distance I move with each keyPress. This is the speed on keyRepeat. Configurable on the class side.

!

!BOBar class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:29'!
color

	^ Color white! !

!BOBar class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:24'!
extentFraction
	"The size as fraction of the playfield"

	^ 5@20! !

!BOBar class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:23'!
speed

	^ 20! !

!BOBar methodsFor: 'initialization' stamp: 'tfel 11/18/2009 09:20'!
initialize

	super initialize.
	self 
		color: self class color;
		speed: self class speed.! !

!BOBar methodsFor: 'event handling' stamp: 'tfel 11/18/2009 09:21'!
moveLeft

	self left: self left - self speed.! !

!BOBar methodsFor: 'event handling' stamp: 'tfel 11/18/2009 09:21'!
moveRight

	self right: self right + self speed.! !

!BOBar methodsFor: 'submorphs-accessing' stamp: 'tfel 11/22/2009 16:04'!
noteNewOwner: aMorph
	
	aMorph isPlayfieldLike ifFalse: [^ self].
	self 
		extent: aMorph extent / self class extentFraction;
		center: (aMorph center x) @ (aMorph bottom - (self height)).! !

!BOBar methodsFor: 'accessing' stamp: 'tfel 11/18/2009 08:21'!
speed
	^ speed! !

!BOBar methodsFor: 'accessing' stamp: 'tfel 11/18/2009 09:21'!
speed: aNumber
	speed := aNumber.! !

RectangleMorph subclass: #BOStone
	instanceVariableNames: 'fieldPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Breakout'!

!BOStone commentStamp: 'tfel 11/22/2009 16:17' prior: 0!
I am a Stone in the playfield, I can be hit by a ball and will the fade away. Also, I should be initialized with newAt:withExtent: and then I will find my own position in the playfield under the assumption that all stones are created with equal extents.

InstanceVariables
	fieldPosition:	Should not be changed from the outside, as this is 
						only relevant at the very time I begin drawing myself 
						in my playfield!

!BOStone class methodsFor: 'configuration' stamp: 'tfel 11/18/2009 08:26'!
color

	^ Color yellow! !

!BOStone class methodsFor: 'configuration' stamp: 'tfel 11/9/2009 16:13'!
fadingSpeed
 
	^ 0.1! !

!BOStone class methodsFor: 'instance creation' stamp: 'tfel 11/19/2009 08:40'!
newAt: aPosition withExtent: aPoint 

	^ self new
		fieldPosition: aPosition;
		extent: aPoint;
		yourself! !

!BOStone methodsFor: 'stepping and presenter' stamp: 'tfel 11/22/2009 15:56'!
fade
	self isStepping ifFalse: [ self startStepping ].
	self color: (self color alpha: self color alpha - self class fadingSpeed).
	self color alpha = 0 ifTrue: [ self abandon ]! !

!BOStone methodsFor: 'accessing' stamp: 'tfel 11/6/2009 19:19'!
fieldPosition
	^ fieldPosition! !

!BOStone methodsFor: 'accessing' stamp: 'tfel 11/18/2009 09:25'!
fieldPosition: aPoint
	fieldPosition := aPoint.! !

!BOStone methodsFor: 'initialization' stamp: 'tfel 11/18/2009 09:25'!
initialize

	super initialize.
	self 
		color: self class color;
		stopStepping.! !

!BOStone methodsFor: 'initialization' stamp: 'tfel 11/22/2009 16:06'!
noteNewOwner: aMorph

	aMorph isPlayfieldLike ifFalse: [^ self].
	self topLeft: (self width@self height * self fieldPosition 
							translateBy: aMorph topLeft).! !

!BOStone methodsFor: 'stepping and presenter' stamp: 'tfel 11/19/2009 08:40'!
step
	self fade! !

!BOStone methodsFor: 'testing' stamp: 'tfel 11/19/2009 08:41'!
stepTime

	^ 40! !

!BOStone methodsFor: 'actions' stamp: 'tfel 11/19/2009 08:41'!
takeHit
	"Extension point: creating differently 'powered' stones"
	self fade! !
